import contextlib
import json
import logging
from enum import IntEnum
from threading import RLock
from typing import (
    Any,
    Callable,
    ContextManager,
    Dict,
    Generic,
    Iterable,
    Mapping,
    MutableMapping,
    Optional,
)

from mnemoreg._storage import MemoeryStorage, StorageProtocol
from mnemoreg._types import K, V
from mnemoreg.exceptions import AlreadyRegisteredError, NotRegisteredError

logger = logging.getLogger(__name__)
logger.setLevel(logging.WARNING)


def locked_method(method: Callable) -> Callable:
    """Decorator to lock method calls for thread safety."""

    def wrapper(self: "Registry", *args: Any, **kwargs: Any) -> Any:
        with self._lock:
            return method(self, *args, **kwargs)

    return wrapper


class OverwritePolicy(IntEnum):
    FORBID = 0
    ALLOW = 1
    WARN = 2


class Registry(MutableMapping, Generic[K, V]):
    """
    Thread-safe registry implementing MutableMapping.

    Arguments:
        lock: Optional lock object to use for synchronization. If None,
            a new RLock is created for threadsafe operation.

    Raises:
        AlreadyRegisteredError: If attempting to register a key that already exists.
        NotRegisteredError: If attempting to access or delete a key that does not exist.

    Examples:
        >>> from typing import Callable
        >>> registry = Registry[str, Callable[[int], int]]()
        >>> registry['a'] = (lambda x: x)  # store a callable
        >>> registry['a'](1)
        1
        >>> @registry.register('b')
        ... def value_b(x: int) -> int:
        ...     return x + 1
        >>> registry['b'](2)
        3
    """

    def __init__(
        self,
        *,
        lock: Optional[RLock] = None,
        log_level: int = logging.WARNING,
        overwrite_policy: int = OverwritePolicy.FORBID,
        store: Optional[StorageProtocol] = None,
    ) -> None:
        # Verify  that lock has the correct methods
        if lock is not None and not all(
            hasattr(lock, method) for method in ("__enter__", "__exit__")
        ):
            raise TypeError("lock must be a threading.RLock or similar object")

        if not (50 >= log_level >= 0):
            raise ValueError("log_level must be a valid logging level between 0 and 50")

        self._lock: RLock = lock or RLock()
        self._store = store or MemoeryStorage()
        self._overwrite_policy = OverwritePolicy(overwrite_policy)
        logger.setLevel(log_level)

    def register(self, key: Optional[K] = None) -> Callable[[V], V]:
        """
        Decorator to register an object with the given key.

        Args:
            key: The key to register the object under. If None, the object's
                 `__name__` attribute will be used.
        Returns:
            A decorator that registers the object and returns it.

        Raises:
            ValueError: If the key is not provided and cannot be inferred
                        from the object.
        """

        def decorator(obj: V) -> V:
            reg_key = key if key is not None else getattr(obj, "__name__", None)
            if reg_key is None:
                raise ValueError(
                    "Registry key must be provided or inferable from object"
                )
            self._validate_key(reg_key, cant_exist=self._overwrite_policy == 0)

            with self._lock:
                # store under the resolved reg_key
                self._store.set(reg_key, obj)
                logger.debug("Registered via decorator %s -> %s", reg_key, type(obj))
            return obj

        return decorator

    def unregister(self, key: K) -> None:
        """Alias for __delitem__ to unregister a key."""
        self.__delitem__(key)

    def remove(self, key: K) -> None:
        """Alias for __delitem__ to remove a key."""
        self.__delitem__(key)

    @locked_method
    def clear(self) -> None:
        self._store.clear()
        logger.debug("Registry cleared")

    @locked_method
    def get(self, key: K, default: Any = None) -> Any:
        return self._store.get(key, default)

    @locked_method
    def snapshot(self) -> Dict[K, V]:
        return self._store.to_dict()

    def to_dict(self) -> Dict[K, V]:
        return self.snapshot()

    @classmethod
    def from_dict(cls, data: Mapping[K, V]) -> "Registry[K, V]":
        r = cls()
        with r._lock:
            r._store.update(dict(data))
        return r

    def to_json(self, **kwargs: Any) -> str:
        return json.dumps(self.to_dict(), **kwargs)

    @classmethod
    def from_json(cls, s: str, **kwargs: Any) -> "Registry[K, V]":
        return cls.from_dict(json.loads(s, **kwargs))

    @locked_method
    def update(self, data: Mapping[K, V]) -> None:
        for k, v in data.items():
            self._validate_key(k, cant_exist=self._overwrite_policy == 0)
            self._store.set(k, v)

    def bulk(self) -> ContextManager["Registry[K, V]"]:
        @contextlib.contextmanager
        def _bulk_ctx():
            self._lock.acquire()
            try:
                yield self
            finally:
                self._lock.release()

        return _bulk_ctx()

    def _validate_key(
        self, key: K, cant_exist: bool = False, must_exist: bool = False
    ) -> None:
        if not isinstance(key, str):
            raise TypeError(f"Registry key must be a string, got {type(key)}")
        elif not key:
            raise ValueError("Registry key cannot be an empty string")
        elif any(c.isspace() for c in key):
            raise ValueError("Registry key cannot contain whitespace characters")
        elif cant_exist and key in self._store.keys():
            raise AlreadyRegisteredError(f"Registry key {key!r} is already registered")
        elif must_exist and key not in self._store.keys():
            raise NotRegisteredError(f"Registry key {key!r} is not registered")

    @locked_method
    def __getitem__(self, key: K) -> V:
        self._validate_key(key, must_exist=True)
        res = self._store.get(key)
        if res is None:
            raise NotRegisteredError(f"Registry key {key!r} is not registered")
        return res

    @locked_method
    def __setitem__(self, key: K, value: V) -> None:
        self._validate_key(key, cant_exist=self._overwrite_policy == 0)
        self._store.set(key, value)
        logger.debug("Registered %s -> %s", key, type(value))

    @locked_method
    def __delitem__(self, key: K) -> None:
        self._validate_key(key, must_exist=True)
        self._store.delete(key)

    @locked_method
    def __iter__(self) -> Iterable[K]:
        return iter(list(self._store.keys()))

    @locked_method
    def __len__(self) -> int:
        return len(self._store)

    @locked_method
    def __contains__(self, key: object) -> bool:
        return key in self._store

    @locked_method
    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({list(self._store.keys())!r})"

    def __getstate__(self):
        return {"_store": self._store.to_dict()}

    def __setstate__(self, state):
        self._lock = RLock()
        self._store.update(state.get("_store", {}))



"""mnemoreg — small thread-safe registry package.

This package exposes a simple `Registry` mapping useful for registering
callables and values by string keys. It's intentionally small and dependency-free.
"""

from importlib.metadata import PackageNotFoundError
from importlib.metadata import version as _pkg_version
from pathlib import Path
from typing import Optional

from mnemoreg._storage import StorageProtocol
from mnemoreg.core import (
    Registry,
)
from mnemoreg.exceptions import AlreadyRegisteredError, NotRegisteredError


def _read_version_file() -> Optional[str]:
    try:
        file_path = Path(__file__).resolve()
        package_dir = file_path.parent
        if package_dir.name == "__pycache__":
            package_dir = package_dir.parent
        version_file = package_dir.joinpath("VERSION")
        return version_file.read_text(encoding="utf8").strip()
    except Exception:
        return None


def _get_version() -> str:
    # 1) Try to read installed distribution metadata
    try:
        return _pkg_version("mnemoreg")
    except PackageNotFoundError:
        pass

    # 2) Try the VERSION file that setuptools_scm can write at build time
    v = _read_version_file()
    if v:
        return v

    # 3) Fall back to a safe default
    return "0.0.0"


__version__ = _get_version()


__all__ = [
    "Registry",
    "AlreadyRegisteredError",
    "NotRegisteredError",
    "__version__",
    "StorageProtocol",
]



import json

import pytest

from mnemoreg import (
    AlreadyRegisteredError,
    NotRegisteredError,
    Registry,
)


def test_set_get_and_delete_basic_behavior():
    r = Registry[str, int]()
    r["a"] = 1
    assert r["a"] == 1
    assert "a" in r
    assert len(r) == 1

    # delete and ensure missing
    del r["a"]
    assert "a" not in r
    assert len(r) == 0
    with pytest.raises(NotRegisteredError):
        _ = r["a"]


def test_duplicate_set_raises():
    r = Registry[str, int]()
    r["x"] = 10
    with pytest.raises(AlreadyRegisteredError):
        r["x"] = 20


def test_delete_not_registered_raises():
    r = Registry[str, int]()
    with pytest.raises(NotRegisteredError):
        del r["nope"]


def test_register_decorator_and_duplicate_decorator():
    r = Registry[str, object]()

    @r.register("f")
    def plus_one(x):
        return x + 1

    assert callable(r["f"])
    assert r.get("f")(3) == 4

    # decorator duplicate should raise at decoration-time
    with pytest.raises(AlreadyRegisteredError):

        @r.register("f")
        def another(x):
            return x


def test_get_default_behavior():
    r = Registry[str, int]()
    assert r.get("missing", "default") == "default"
    assert r.get("missing") is None


def test_snapshot_and_to_dict_and_from_dict_are_shallow_and_independent():
    r = Registry[str, dict]()
    r["conf"] = {"a": 1}

    snap = r.snapshot()
    assert snap == {"conf": {"a": 1}}

    # mutate snapshot and ensure original isn't affected
    snap["conf"]["a"] = 999  # shallow change: underlying value is same reference
    assert r["conf"]["a"] == 999  # values are same reference (shallow copy)
    # but replacing mapping in snapshot doesn't change original
    d = r.to_dict()
    d["conf"] = {"a": 0}
    assert r["conf"]["a"] == 999

    new = Registry.from_dict({"conf": {"a": 2}})
    assert new["conf"]["a"] == 2


def test_to_json_and_from_json_roundtrip():
    r = Registry[str, int]()
    r["num"] = 42
    s = r.to_json()
    # ensure it's valid JSON and roundtrips
    obj = json.loads(s)
    assert obj == {"num": 42}

    new = Registry.from_json(s)
    assert new["num"] == 42


def test_iter_returns_snapshot_iterator():
    r = Registry[str, int]()
    r["k1"] = 1
    it = iter(r)
    # mutate registry after getting iterator
    r["k2"] = 2
    # iterator was created from snapshot: should only contain k1
    items = list(it)
    assert items == ["k1"]
    # but iterating anew sees both
    assert list(iter(r)) == ["k1", "k2"]


def test_repr_contains_keys_and_len_and_contains_non_str_key():
    r = Registry[str, int]()
    r["one"] = 1
    r["two"] = 2
    rep = repr(r)
    assert "one" in rep and "two" in rep
    # __contains__ should accept any object (false for non-existing)
    assert (123 in r) is False


def test_bulk_context_manager_and_exception_propagation_releases_lock():
    r = Registry[str, int]()

    # normal usage should allow modifications while holding lock
    with r.bulk() as reg:
        # we are inside the context, it's the same object
        assert reg is r
        reg["a"] = 1
        assert r["a"] == 1

    # exception inside must propagate (return False from __exit__)
    # and not leave lock acquired
    with pytest.raises(ValueError):
        with r.bulk() as reg:
            reg["b"] = 2
            raise ValueError("boom")

    # after the exception, we can re-enter bulk (meaning lock was released)
    with r.bulk() as reg2:
        reg2["c"] = 3
        assert r["c"] == 3



from abc import ABC, abstractmethod
from typing import Dict, Generic, Iterator, Mapping, Optional

from mnemoreg._types import K, V


class AbstractStorage(ABC, Generic[K, V]):
    """Abstract base class for storage implementations."""

    @abstractmethod
    def set(self, key: K, value: V) -> None:
        pass

    @abstractmethod
    def get(self, key: K, default: Optional[V] = None) -> Optional[V]:
        pass

    @abstractmethod
    def delete(self, key: K) -> None:
        pass

    @abstractmethod
    def clear(self) -> None:
        pass

    @abstractmethod
    def to_dict(self) -> Dict[K, V]:
        pass

    @abstractmethod
    def update(self, data: Mapping[K, V]) -> None:
        pass

    @abstractmethod
    def keys(self) -> Iterator[K]:
        pass

    @abstractmethod
    def __len__(self) -> int:
        pass

    @abstractmethod
    def __contains__(self, key: object) -> bool:
        pass



import json

import pytest

from mnemoreg import AlreadyRegisteredError, NotRegisteredError, Registry


def test_access_before_registration_raises():
    r = Registry[str, int]()
    with pytest.raises(NotRegisteredError):
        _ = r["missing"]


def test_reassign_same_key_is_not_allowed():
    r = Registry[str, int]()
    r["x"] = 1
    with pytest.raises(AlreadyRegisteredError):
        r["x"] = 2


def test_delete_twice_raises_cleanly():
    r = Registry[str, int]()
    r["a"] = 1
    del r["a"]
    with pytest.raises(NotRegisteredError):
        del r["a"]


def test_register_decorator_duplicate_key_fails():
    r = Registry[str, object]()

    @r.register("f")
    def f(x):
        return x

    with pytest.raises(AlreadyRegisteredError):

        @r.register("f")
        def g(y):
            return y


def test_register_with_non_string_key_type_hint_violation():
    r = Registry[str, int]()
    # Type hint violation—should still technically work but we test safety expectations
    r["1"] = 1
    with pytest.raises(TypeError):
        _ = r[1]  # using int instead of str


def test_get_returns_default_instead_of_raising():
    r = Registry[str, int]()
    assert r.get("unknown", 999) == 999
    assert r.get("unknown") is None


def test_to_json_fails_on_non_serializable_content():
    class Dummy:
        pass

    r = Registry[str, object]()
    r["obj"] = Dummy()
    with pytest.raises(TypeError):
        r.to_json()


def test_from_json_with_invalid_json_string():
    with pytest.raises(json.JSONDecodeError):
        Registry.from_json("{invalid}")


def test_snapshot_cannot_mutate_original_store_keys_directly():
    r = Registry[str, list]()
    r["x"] = [1, 2, 3]
    snap = r.snapshot()
    snap["x"].append(4)  # same reference value
    assert r["x"] == [1, 2, 3, 4]
    # but removing a key in snapshot has no effect on registry
    del snap["x"]
    assert "x" in r


def test_bulk_context_does_not_swallow_exceptions():
    r = Registry[str, int]()
    with pytest.raises(ValueError):
        with r.bulk():
            r["a"] = 1
            raise ValueError("force exit")
    # confirm lock released
    r["b"] = 2
    assert r["b"] == 2



from typing import Any, Dict, Generic, Iterator, Mapping, Protocol, runtime_checkable

from mnemoreg._types import K, V


@runtime_checkable
class StorageProtocol(Protocol, Generic[K, V]):
    """Minimal protocol describing the storage interface expected by Registry.

    Only the methods and members that `mnemoreg.core.Registry` uses are
    specified here so the protocol stays small and permissive.
    """

    def set(self, key: K, value: V) -> None:  # pragma: no cover - interface
        ...

    def get(self, key: K, default: Any = None) -> Any:  # pragma: no cover - interface
        ...

    def delete(self, key: K) -> None:  # pragma: no cover - interface
        ...

    def clear(self) -> None:  # pragma: no cover - interface
        ...

    def to_dict(self) -> Dict[K, V]:  # pragma: no cover - interface
        ...

    def update(self, data: Mapping[K, V]) -> None:  # pragma: no cover - interface
        ...

    def keys(self) -> Iterator[K]:  # pragma: no cover - interface
        ...

    def __len__(self) -> int:  # pragma: no cover - interface
        ...

    def __contains__(self, key: object) -> bool:  # pragma: no cover - interface
        ...



import threading

import pytest

from mnemoreg import AlreadyRegisteredError, NotRegisteredError, Registry

pytestmark = pytest.mark.filterwarnings(
    "ignore::pytest.PytestUnhandledThreadExceptionWarning"
)


def _writer(reg: Registry, start: int, end: int):
    for i in range(start, end):
        key = f"k{i}"
        reg[key] = i


def _reader(reg: Registry, start: int, end: int, results: list):
    for i in range(start, end):
        key = f"k{i}"
        try:
            val = reg.get(key, None)
            if val is not None:
                results.append(val)
        except Exception as e:
            results.append(str(e))


def _deleter(reg: Registry):
    for i in range(0, 100, 2):
        try:
            del reg[f"k{i}"]
        except NotRegisteredError:
            pass


def _getter(reg: Registry, results: list):
    for i in range(100):
        try:
            val = reg.get(f"k{i}")
            results.append(val)
        except NotRegisteredError:
            results.append(None)


def _make_and_register_func(reg: Registry, n: int):
    @reg.register(f"f{n}")
    def f(x):
        return x + n

    return f


def _bulk_writer(reg: Registry, start: int, end: int):
    with reg.bulk() as r:
        for i in range(start, end):
            key = f"k{i}"
            if key not in r:
                r[key] = i
            else:
                try:
                    r[key] = i
                except AlreadyRegisteredError:
                    pass


def _mutator(reg: Registry):
    for i in range(50, 100):
        reg[f"k{i}"] = i


def _snapper(reg: Registry, out_list: list):
    out_list.append(reg.snapshot())


def test_concurrent_set_and_get():
    r = Registry[str, int]()

    write_threads = [
        threading.Thread(target=_writer, args=(r, i * 10, (i + 1) * 10))
        for i in range(5)
    ]
    read_results = []
    read_threads = [
        threading.Thread(target=_reader, args=(r, 0, 50, read_results))
        for _ in range(5)
    ]

    # Start writers
    for t in write_threads:
        t.start()
    # Start readers concurrently
    for t in read_threads:
        t.start()

    for t in write_threads + read_threads:
        t.join()

    # All keys should exist
    for i in range(50):
        key = f"k{i}"
        assert key in r
        assert r[key] == i
    # Reader results contain only valid values
    for val in read_results:
        assert isinstance(val, int)


def test_concurrent_deletion_and_access():
    r = Registry[str, int]()
    for i in range(100):
        r[f"k{i}"] = i

    threads = []
    results = []

    for _ in range(5):
        threads.append(threading.Thread(target=_deleter, args=(r,)))
        threads.append(threading.Thread(target=_getter, args=(r, results)))

    for t in threads:
        t.start()
    for t in threads:
        t.join()

    # Check all remaining keys are odd
    for i in range(0, 100, 2):
        assert f"k{i}" not in r
    for i in range(1, 100, 2):
        assert f"k{i}" in r
        assert r[f"k{i}"] == i


def test_concurrent_register_decorator():
    r = Registry[str, int]()

    threads = [
        threading.Thread(
            target=_make_and_register_func,
            args=(
                r,
                i,
            ),
        )
        for i in range(20)
    ]
    for t in threads:
        t.start()
    for t in threads:
        t.join()

    # All registered functions must work
    for i in range(20):
        f = r[f"f{i}"]
        assert f(10) == 10 + i


def test_bulk_context_under_concurrency():
    r = Registry[str, int]()

    threads = [
        threading.Thread(target=_bulk_writer, args=(r, i * 10, (i + 1) * 10))
        for i in range(10)
    ]
    for t in threads:
        t.start()
    for t in threads:
        t.join()

    # All keys should exist
    for i in range(100):
        key = f"k{i}"
        assert key in r
        assert isinstance(r[key], int)


def test_snapshot_under_concurrent_modification():
    r = Registry[str, int]()

    for i in range(50):
        r[f"k{i}"] = i

    snapshot_results = []

    threads = []
    for _ in range(5):
        threads.append(threading.Thread(target=_mutator, args=(r,)))
        threads.append(threading.Thread(target=_snapper, args=(r, snapshot_results)))

    for t in threads:
        t.start()
    for t in threads:
        t.join()

    # All snapshots should be subsets of final state
    final_keys = set(r)
    for snap in snapshot_results:
        assert set(snap.keys()).issubset(final_keys)
